--------------------------------------------------------------------------------------------------
-- script : sync-packages.lua
-- author : Dan Gill
-- June 2017
-- version: 1.5
---------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------
-- desc   : The intention of this script is to automate the process of updating packages. 
-- This script should be run by the nas with arguments passed to it. The first argument must be the 
-- package name, the second argument must be the version, and the third argument must be whether a
-- fresh install is desired (deploy regardless of whether probe currently is installed on the robot),
-- the fourth argument is the hub name ro process or ALL for all hubs.
-- This script will EXCLUDE ALL HUB ROBOTS from processing.
-- Alternatively, the script may be run manually and will use the variables provided in the script.
-- The folder "script_logs" *MUST* be manually created under the nas directory prior to running this
-- script.
---------------------------------------------------------------------------------------------------        
---------------------------------------------------------------------------------------------------
-- Version  | Details
---------------------------------------------------------------------------------------------------
-- 1.0      | Initial Version
-- 1.1      | 9/13/2018 Added parameter to specify if script should be run against a single hub or
--            ALL. Converted to used more functions for faster processing. Converted output to file.
-- 1.2      | 9/20/2018 - Replace dofile with require and added package.path update; this makes
--            passing the path to the includes files simpler. Swapped out file.write and print
--            statements with output function from logging_functions.
-- 1.3      | 10/3/2018 - Added additional error checking.
-- 1.4      | 11/13/2018 - Added hub name to output log. Also added task dump to log output to ease
--            probe callback get_task_status against the ADE.
-- 1.5      | 11/14/2018 - Added PUSH, UPDATE, and INSTALL installation types for various types of
--            package installs
---------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------
---------------------------------------------------------------------------------------------------

-----------------------------------------------------------------------
-----------------------------------------------------------------------
-- Set Variables
-----------------------------------------------------------------------
-----------------------------------------------------------------------

package.path = package.path .. ";./scripts/includes/?.lua" -- Where are your require files?

local base_pkg_name = "_ntservices_default_monitoring" -- The name of the base probe package
local base_pkg_ver  = "1.2" -- The version of the base probe package
-- PUSH = Push package to all robots; INSTALL = Install if not already installed and/or not at
-- correct version; UPDATE = Ensure package version is correct if package is already installed.
local install_type = "INSTALL"  
local hub = "mev-nshub01" -- Specify hub (hub name only, full path isn't needed) or ALL
local output_location = 2 -- 1 = stdout; 2 = file; 3 = both

-----------------------------------------------------------------------
-----------------------------------------------------------------------
-- DO NOT EDIT BELOW THIS LINE
-----------------------------------------------------------------------
-----------------------------------------------------------------------

require ("logging_functions")
require ("error_functions")
require ("table_functions")

if SCRIPT_ARGUMENT ~= nil then
   parms = split(SCRIPT_ARGUMENT)
   for k,v in ipairs (parms) do
      if k == 1 then
         base_pkg_name = v -- The name of the base probe package
      elseif k == 2 then
         base_pkg_ver = v -- The version of the base probe package
      elseif k == 3 then
         install_type = v -- 0 = Update only; 1 = Fresh Install
      elseif k == 4 then
         hub = v
      end
   end
end

local str_beg, str_end = string.find (SCRIPT_NAME,".",1,true)
local fname = "./script_logs/" .. left (SCRIPT_NAME, str_beg-1) .. "-" .. base_pkg_name .. "-" .. hub .. ".log"

----------------------------------------------------------------------
-----------------------------------------------------------------------
-- Get UIM robot details
-----------------------------------------------------------------------
-----------------------------------------------------------------------


-- Return a list of all hubs under the domain
local function get_hublist()

   local hubs, rc = nimbus.request("hub", "gethubs")
   if rc == 0 then
      return hubs.hublist, rc
   else
      output(fname, timestamp.format ( timestamp.now(), "%Y-%m-%d %H:%M:%S").." Error running gethubs callback against primary hub.", output_location)
      codes_file(rc,fname,output_location)
      return nil, rc
   end
end

-- Return a list of all robots under a hub
local function get_robotlist(hub_addr)

   local robots, rc = nimbus.request(hub_addr, "getrobots")
   if rc == 0 then
      return robots.robotlist, rc
   else
      output(fname, timestamp.format ( timestamp.now(), "%Y-%m-%d %H:%M:%S").." Error running getrobots callback against hub: " .. hub_addr, output_location)
      codes_file(rc,fname,output_location)
      return nil, rc
   end
  
end

-----------------------------------------------------------------------
-----------------------------------------------------------------------
-- Deploy base package
-----------------------------------------------------------------------
-----------------------------------------------------------------------
local function deploy_base_package(pkg_name, pkg_ver, robots, hubrobot)
   local job = pds.create ()
   local i = 0
   pds.putString (job, "jobname", pkg_name.."-"..pkg_ver .. "-" .. hubrobot)

   for key,_ in pairs (robots) do
      if hubrobot ~= robots[key].name then -- Do not run the script against hubs
--      if robots[key].name == "TEST" then -- Test against a single robot    THIS LINE OR THE LINE ABOVE MUST BE COMMENTED AND THE OTHER UNCOMMENTED
         local package_args = pds.create()
         pds.putString (package_args, "package", pkg_name)
         local inst_list, ret = nimbus.request(robots[key].addr .. "/controller", "inst_list", package_args)
         pds.delete(package_args)
         if ret == 0 then
            if install_type == "PUSH" or (install_type == "INSTALL" and inst_list[pkg_name] == nil) then
                  local args = pds.create()
                  pds.putString (args, "package", pkg_name)
                  pds.putString (args, "version", pkg_ver)
                  pds.putString (args, "robot", robots[key].addr)
--                  pds.putString (args, "update", fresh_install) -- Documentation is unclear https://search.ca.com/assets/SiteAssets/TEC1453101_External/Transitioning%20from%20Distsrv%20to%20ADE%202.0.pdf
                  pds.putTable (job, "probes", args)
                  pds.delete(args)            
            elseif inst_list[pkg_name] ~= nil and (install_type == "UPDATE" or install_type == "INSTALL") then
               if inst_list[pkg_name].version ~= pkg_ver then
                  local args = pds.create()
                  pds.putString (args, "package", pkg_name)
                  pds.putString (args, "version", pkg_ver)
                  pds.putString (args, "robot", robots[key].addr)
--                  pds.putString (args, "update", fresh_install) -- Documentation is unclear https://search.ca.com/assets/SiteAssets/TEC1453101_External/Transitioning%20from%20Distsrv%20to%20ADE%202.0.pdf
                  pds.putTable (job, "probes", args)
                  pds.delete(args)
               end
            end
         else
            output(fname, timestamp.format ( timestamp.now(), "%Y-%m-%d %H:%M:%S").." Error running callback against robot: " .. robots[key].addr, output_location)
            codes_file(ret,fname,output_location)
         end
         i = i+1
      end
   end

--   tdump_file(pds.convert (job), fname, output_location)

   if i > 0 then
      local job_id, ret = nimbus.request("automated_deployment_engine","submit_job", job)
--   codes(ret)

      if ret == 0 and job_id ~= nil then
         output(fname, "\n" .. timestamp.format ( timestamp.now(), "%Y-%m-%d %H:%M:%S").." Package deployment queued on " .. hubrobot .. ". Job ID " .. job_id.JobID, output_location)
         -- local args = pds.create()
         -- pds.putString (args, "JobID", job_id.JobID)
         -- local tasks, tasks_rc = nimbus.request("automated_deployment_engine", "get_job_summary", args)
         -- pds.delete(args)
         -- tdump_file(pds.convert (tasks), fname, output_location)
      elseif ret ~= nil then
         output(fname, "\n" .. timestamp.format ( timestamp.now(), "%Y-%m-%d %H:%M:%S").." Error deploying package.", output_location)
         codes_file(ret, fname, output_location)
      else
         output(fname, "\n" .. timestamp.format ( timestamp.now(), "%Y-%m-%d %H:%M:%S").." Error deploying package, no return code received.", output_location)
      end
   else
      output(fname, "\n" .. timestamp.format ( timestamp.now(), "%Y-%m-%d %H:%M:%S").." No robots match criteria for package distribution.", output_location)
   end
   pds.delete(job)
end

local function main()
   -- Creates/overwrites the file - only do once
   if output_location == 2 or output_location == 3 then
      file.create (fname)
   end

   output(fname,timestamp.format ( timestamp.now(), "%Y-%m-%d %H:%M:%S").." Deploying package " .. base_pkg_name .. ", version: " .. base_pkg_ver .." to hub(s): " .. hub ..  ". Install  Type: " .. install_type, output_location)

   local hublist, rc_hubs = get_hublist()

   if rc_hubs == 0 then
      for k,v in pairs (hublist) do
         if hub == hublist[k].name or hub == "ALL" then
            local robotlist, rc_robots = get_robotlist(hublist[k].addr)
            if rc_robots == 0 then
               deploy_base_package(base_pkg_name, base_pkg_ver, robotlist, hublist[k].robotname)
            end
         end
      end
   end
   
   output(fname, timestamp.format ( timestamp.now(), "%Y-%m-%d %H:%M:%S").." Script has completed running.", output_location)
   
end

main()